from typing import Optional

from flask import Response

from ieee_2030_5.server.base_request import RequestOp


class DERProgramRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self, edev_id: Optional[int] = None, id: Optional[int] = None) -> Response:
        return Response("Foo")



from __future__ import annotations

import logging
import threading
from http.client import HTTPSConnection
from os import PathLike
from pathlib import Path
import ssl
import atexit
from typing import Optional, Dict, Tuple
import xml.dom.minidom
from threading import Timer

import xsdata

from ieee_2030_5.models import DeviceCapability, EndDeviceListLink, MirrorUsagePointList, MirrorUsagePoint, \
    UsagePointList, EndDevice, Registration, FunctionSetAssignmentsListLink, Time, DERProgramList, \
    FunctionSetAssignments

from ieee_2030_5.utils import dataclass_to_xml, parse_xml

_log = logging.getLogger(__name__)


class IEEE2030_5_Client:
    clients: set[IEEE2030_5_Client] = set()

    # noinspection PyUnresolvedReferences
    def __init__(self,
                 cafile: Path,
                 server_hostname: str,
                 keyfile: Path,
                 certfile: Path,
                 server_ssl_port: Optional[int] = 443,
                 debug: bool = True):

        assert cafile.exists(), f"cafile doesn't exist ({cafile})"
        assert keyfile.exists(), f"keyfile doesn't exist ({keyfile})"
        assert certfile.exists(), f"certfile doesn't exist ({certfile})"

        self._ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        self._ssl_context.verify_mode = ssl.CERT_REQUIRED
        self._ssl_context.load_verify_locations(cafile=cafile)

        # Loads client information from the passed cert and key files. For
        # client side validation.
        self._ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)

        self._http_conn = HTTPSConnection(host=server_hostname,
                                          port=server_ssl_port,
                                          context=self._ssl_context)
        self._device_cap: Optional[DeviceCapability] = None
        self._mup: Optional[MirrorUsagePointList] = None
        self._upt: Optional[UsagePointList] = None
        self._edev: Optional[EndDeviceListLink] = None
        self._end_devices: Optional[EndDeviceListLink] = None
        self._fsa_list: Optional[FunctionSetAssignmentsListLink] = None
        self._debug = debug
        self._dcap_poll_rate: int = 0
        self._dcap_timer: Optional[Timer] = None
        self._disconnect: bool = False

        IEEE2030_5_Client.clients.add(self)

    @property
    def http_conn(self) -> HTTPSConnection:
        if self._http_conn.sock is None:
            self._http_conn.connect()
        return self._http_conn

    def new_uuid(self, url: str = "/uuid") -> str:
        res = self.__get_request__(url)
        return res

    def end_devices(self) -> EndDeviceListLink:
        self._end_devices = self.__get_request__(self._device_cap.EndDeviceListLink.href)
        return self._end_devices

    def end_device(self, index: Optional[int] = 0) -> EndDevice:
        if not self._end_devices:
            self.end_devices()

        return self._end_devices.EndDevice[index]

    def self_device(self) -> EndDevice:
        if not self._device_cap:
            self.device_capability()

        return self.__get_request__(self._device_cap.SelfDeviceLink.href)

    def function_set_assignment(self) -> FunctionSetAssignmentsListLink:
        fsa_list = self.__get_request__(self.self_device().FunctionSetAssignmentsListLink.href)
        return fsa_list

    def poll_timer(self, fn, args):
        if not self._disconnect:
            _log.debug(threading.currentThread().name)
            fn(args)
            threading.currentThread().join()

    def device_capability(self, url: str = "/dcap") -> DeviceCapability:
        self._device_cap: DeviceCapability = self.__get_request__(url)
        if self._device_cap.pollRate is not None:
            self._dcap_poll_rate = self._device_cap.pollRate
        else:
            self._dcap_poll_rate = 600

        _log.debug(f"devcap id {id(self._device_cap)}")
        _log.debug(threading.currentThread().name)
        _log.debug(f"DCAP: Poll rate: {self._dcap_poll_rate}")
        self._dcap_timer = Timer(self._dcap_poll_rate, self.poll_timer, (self.device_capability, url))
        self._dcap_timer.start()
        return self._device_cap

    def time(self) -> Time:
        timexml = self.__get_request__(self._device_cap.TimeLink.href)
        return timexml

    def der_program_list(self, device: EndDevice) -> DERProgramList:
        fsa: FunctionSetAssignments = self.__get_request__(device.FunctionSetAssignmentsListLink.href)
        der_programs_list: DERProgramList = self.__get_request__(fsa.DERProgramListLink.href)

        return der_programs_list

    def mirror_usage_point_list(self) -> MirrorUsagePointList:
        self._mup = self.__get_request__(self._device_cap.MirrorUsagePointListLink.href)
        return self._mup

    def usage_point_list(self) -> UsagePointList:
        self._upt = self.__get_request__(self._device_cap.UsagePointListLink.href)
        return self._upt

    def registration(self, end_device: EndDevice) -> Registration:
        reg = self.__get_request__(end_device.RegistrationLink.href)
        return reg

    def timelink(self):
        if self._device_cap is None:
            raise ValueError("Request device capability first")
        return self.__get_request__(url=self._device_cap.TimeLink.href)

    def disconnect(self):
        self._disconnect = True
        self._dcap_timer.cancel()
        IEEE2030_5_Client.clients.remove(self)

    def request(self, endpoint: str, body: dict = None, method: str = "GET",
                headers: dict = None):

        if method.upper() == 'GET':
            return self.__get_request__(endpoint, body, headers=headers)

        if method.upper() == 'POST':
            print("Doing post")
            return self.__post__(endpoint, body, headers=headers)

    def create_mirror_usage_point(self, mirror_usage_point: MirrorUsagePoint) -> Tuple[int, str]:
        data = dataclass_to_xml(mirror_usage_point)
        resp = self.__post__(self._device_cap.MirrorUsagePointListLink.href, data=data)
        return resp.status, resp.headers['Location']

    def __post__(self, url: str, data=None, headers: Optional[Dict[str, str]]=None):
        if not headers:
            headers = {'Content-Type': 'text/xml'}

        self.http_conn.request(method="POST", headers=headers,
                               url=url, body=data)
        response = self._http_conn.getresponse()
        # response_data = response.read().decode("utf-8")

        return response

    def __get_request__(self, url: str, body=None, headers: dict = None):
        if headers is None:
            headers = {"Connection": "keep-alive", "keep-alive": "timeout=30, max=1000"}

        if self._debug:
            print(f"----> GET REQUEST")
            print(f"url: {url} body: {body}")
        self.http_conn.request(method="GET", url=url, body=body, headers=headers)
        response = self._http_conn.getresponse()
        response_data = response.read().decode("utf-8")
        print(response.headers)

        response_obj = None
        try:
            response_obj = parse_xml(response_data)
            resp_xml = xml.dom.minidom.parseString(response_data)
            if resp_xml and self._debug:
                print(f"<---- GET RESPONSE")
                print(f"{response_data}")  # toprettyxml()}")

        except xsdata.exceptions.ParserError as ex:
            if self._debug:
                print(f"<---- GET RESPONSE")
                print(f"{response_data}")
            response_obj = response_data

        return response_obj

    def __close__(self):
        self._http_conn.close()
        self._ssl_context = None
        self._http_conn = None


# noinspection PyTypeChecker
def __release_clients__():
    for x in IEEE2030_5_Client.clients:
        x.__close__()
    IEEE2030_5_Client.clients = None


atexit.register(__release_clients__)

#
# ssl_context = ssl.create_default_context(cafile=str(SERVER_CA_CERT))
#
#
# con = HTTPSConnection("me.com", 8000,
#                       key_file=str(KEY_FILE),
#                       cert_file=str(CERT_FILE),
#                       context=ssl_context)
# con.request("GET", "/dcap")
# print(con.getresponse().read())
# con.close()

if __name__ == '__main__':
    SERVER_CA_CERT = Path("~/tls/certs/ca.crt").expanduser().resolve()
    KEY_FILE = Path("~/tls/private/_def62366-746e-4fcb-b3ee-ebebb90d72d4.pem").expanduser().resolve()
    CERT_FILE = Path("~/tls/certs/_def62366-746e-4fcb-b3ee-ebebb90d72d4.crt").expanduser().resolve()

    headers = {'Connection': 'Keep-Alive',
               'Keep-Alive': "max=1000,timeout=30"}

    h = IEEE2030_5_Client(cafile=SERVER_CA_CERT,
                          server_hostname="gridappsd_dev_2004",
                          server_ssl_port=8443,
                          keyfile=KEY_FILE,
                          certfile=CERT_FILE,
                          hostname='_def62366-746e-4fcb-b3ee-ebebb90d72d4')
    # h2 = IEEE2030_5_Client(cafile=SERVER_CA_CERT, server_hostname="me.com", ssl_port=8000,
    #                        keyfile=KEY_FILE, certfile=KEY_FILE)
    resp = h.request("/dcap", headers=headers)
    print(resp)
    resp = h.request("/dcap", headers=headers)
    print(resp)
    #dcap = h.device_capability()
    # get device list
    #dev_list = h.request(dcap.EndDeviceListLink.href).EndDevice

    #ed = h.request(dev_list[0].href)
    #print(ed)
    #
    # print(dcap.mirror_usage_point_list_link)
    # # print(h.request(dcap.mirror_usage_point_list_link.href))
    # print(h.request("/dcap", method="post"))


    # tl = h.timelink()
    #print(IEEE2030_5_Client.clients)



import json

from flask import Flask, Response, render_template

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.server.server_constructs import EndDevices


class AdminEndpoints:
    def __init__(self, app: Flask, end_devices: EndDevices, tls_repo: TLSRepository, config: ServerConfiguration):
        self.end_devices = end_devices
        self.tls_repo = tls_repo
        self.server_config = config

        app.add_url_rule("/admin/end-device-list", view_func=self._end_device_list)
        app.add_url_rule("/admin/program-lists", view_func=self._program_lists)
        app.add_url_rule("/admin/lfid", endpoint="admin/lfid", view_func=self._lfid_lists)
        app.add_url_rule("/admin/edev/<int:edevid>/fsa", view_func=self._edev_fsa)

    def _lfid_lists(self) -> Response:
        items = []

        for k, v in self.end_devices.__all_end_devices__.items():
            items.append({"key": k, "lfid": int(v.end_device.lFDI)})

        return Response(json.dumps(items))

    def _edev_fsa(self, edevid: int) -> Response:
        #edev = self.end_devices.get(edevid)
        return Response(json.dumps(json.dumps(self.end_devices.get_fsa_list(edevid=edevid))))

    def _program_lists(self) -> str:
        return render_template("admin/program-lists.html",
                               program_lists=self.server_config.program_lists)

    def _end_device_list(self) -> str:
        return render_template("admin/end-device-list.html",
                               end_device_list=self.end_devices.get_end_devices())



from typing import Optional, List

DEFAULT_DCAP_ROOT = "/dcap"
DEFAULT_EDEV_ROOT = "/edev"
DEFAULT_UPT_ROOT = "/utp"
DEFAULT_MUP_ROOT = "/mup"
DEFAULT_DRP_ROOT = "/drp"
DEFAULT_SELF_ROOT = "/sdev"
DEFAULT_MESSAGE_ROOT = "/msg"


dcap: str = f"{DEFAULT_DCAP_ROOT}"
# TimeLink
tm: str = f"{DEFAULT_DCAP_ROOT}/tm"
# ResponseSetListLink
rsps: str = f"{DEFAULT_DCAP_ROOT}/rsps"
# UsagePointListLink
upt: str = DEFAULT_UPT_ROOT

# DERProgramListLink
derp: str = "/derp"

# EndDeviceListLink
edev: str = DEFAULT_EDEV_ROOT
edev_urls: List = [
    edev,
    f"{edev}/<int:index>",
    f"{edev}/<int:index>/<category>",
    f"{edev}/<int:index>/<category>/<sub_index>"
]

# MirrorUsagePointListLink
mup: str = DEFAULT_MUP_ROOT
mup_urls: List = [
    (mup, ('GET', 'POST')),
    f"{mup}/<int:index>"
]

curve: str = "/curves"
curve_urls: List = [
    f"{curve}",
    (f"{curve}/<int:index>", ("GET",))
]

program: str = "/programs"
program_urls: List = [
    f"{program}",
    (f"{program}/<int:index>/actderc", ("GET",)),
    (f"{program}/<int:index>/dc", ("GET",)),
    (f"{program}/<int:index>/dderc", ("GET",)),
    (f"{program}/<int:index>/derc", ("GET",)),
]

der: str = "/der"
der_urls: List = [
    (f"{der}/<int:edev_id>", ('GET', 'POST')),
    (f"{der}/<int:edev_id>/<int:id>", ('GET', 'PUT', 'DELETE')),
    (f"{der}/<int:edev_id>/<int:id>/upt", ('GET', 'DELETE')),
    (f"{der}/<int:edev_id>/<int:id>/derp", ('GET', 'POST')),
    (f"{der}/<int:edev_id>/<int:id>/cdp", ('GET', 'DELETE')),
    (f"{der}/<int:edev_id>/<int:id>/derg", ('GET', 'PUT')),
    (f"{der}/<int:edev_id>/<int:id>/ders", ('GET', 'PUT')),
    (f"{der}/<int:edev_id>/<int:id>/dera", ('GET', 'PUT')),
    (f"{der}/<int:edev_id>/<int:id>/dercap", ('GET', 'PUT')),
]

sdev: str = DEFAULT_SELF_ROOT


def build_der_link(edev_id: Optional[int] = None, id: Optional[int] = None, suffix: Optional[str] = None) -> str:
    if edev_id is None:
        raise ValueError("edev_id must be specified.")
    if id is not None and suffix is not None:
        link = build_link(f"{der}", f"{edev_id}", f"{id}", suffix)
    elif id is not None:
        link = build_link(f"{der}", f"{edev_id}", f"{id}")
    elif suffix is not None:
        link = build_link(f"{der}", f"{edev_id}", suffix)
    else:
        link = build_link(f"{der}", f"{edev_id}")

    return link


def build_edev_registration_link(index: int) -> str:
    return build_link(f"{edev}", index, "reg")


def build_edev_status_link(index: int) -> str:
    return build_link(f"{edev}", index, "ds")


def build_edev_config_link(index: int) -> str:
    return build_link(f"{edev}", index, "cfg")


def build_edev_info_link(index: int) -> str:
    return build_link(f"{edev}", index, "di")


def build_edev_power_status_link(index: int) -> str:
    return build_link(f"{edev}", index, "ps")


def build_edev_fsa_link(index: int, fsa_index: Optional[int] = None) -> str:
    return build_link(f"{edev}", index, "fsa", fsa_index)


# edev_cfg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/cfg"
# edev_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ds"
# edev_info_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# edev_power_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
# edev_file_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fs"
# edev_sub_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/subl"

#
# # DemandResponseProgramListLink
# drp: str = DEFAULT_DRP_ROOT
# # MessagingProgramListLink
# msg: str = DEFAULT_MESSAGE_ROOT
# # SelfDeviceLink
# sdev: str = DEFAULT_SELF_ROOT
#
# edev_base: str = '/edev'
# edev: List[str] = [
#     DEFAULT_EDEV_ROOT,
#     [f"{DEFAULT_EDEV_ROOT}/<int:index>", ["GET", "POST"]]
# ]
#
#
#
# sdev_di: str = f"{DEFAULT_DCAP_ROOT}/sdev/di"
# sdev_log: str = f"{DEFAULT_DCAP_ROOT}/sdev/log"
#
# mup_fmt: str = f"{DEFAULT_MUP_ROOT}" + "/{index}"
#
# edev_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}"
# reg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/reg"
# # di_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# #dstat_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dstat"
# ps_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
#
# derp_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derp"
# derp_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derp/1"
#
# der_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/der/1"
# dera_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dera/1"
# dercap_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/dercap/1"
# derg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derg/1"
# ders_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ders/1"
#
# derc_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derc"
# derc_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/derc/1"
#
# fsa_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fsa"
# fsa_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fsa/0"
#
# edev_cfg_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/cfg"
# edev_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ds"
# edev_info_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/di"
# edev_power_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/ps"
# edev_file_status_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/fs"
# edev_sub_list_fmt: str = f"{DEFAULT_DCAP_ROOT}/edev" + "/{index}/subl"

admin: str = "/admin"
uuid_gen: str = "/uuid"


def build_link(base_url: str, *suffix: Optional[str]):
    result = base_url
    if result.endswith("/"):
        result = result[:-1]

    if suffix:
        for p in suffix:
            if p is not None:
                if isinstance(p, str):
                    if p.startswith("/"):
                        result += f"{p}"
                    else:
                        result += f"/{p}"
                else:
                    result += f"/{p}"

    return result


def extend_url(base_url: str, index: Optional[int] = None, suffix: Optional[str] = None):
    result = base_url
    if index is not None:
        result += f"/{index}"
    if suffix:
        result += f"/{suffix}"

    return result



from dataclasses import dataclass
from typing import Type, Optional

from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers.config import ParserConfig
from xsdata.formats.dataclass.parsers.xml import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

__xml_context__ = XmlContext()
__parser_config__ = ParserConfig(fail_on_unknown_attributes=False,
                                 fail_on_unknown_properties=False)
__xml_parser__ = XmlParser(config=__parser_config__, context=__xml_context__)
__config__ = SerializerConfig(xml_declaration=False, pretty_print=True)
__serializer__ = XmlSerializer(config=__config__)
__ns_map__ = {None: "urn:ieee:std:2030.5:ns"}


def serialize_dataclass(obj: dataclass) -> str:
    """
    Serializes a dataclass that was created via xsdata to an xml string for
    returning to a client.
    """
    return __serializer__.render(obj, ns_map=__ns_map__)


def parse_xml(xml: str, type: Optional[Type] = None) -> dataclass:
    """
    Parse the xml passed and return result from loaded classes.
    """
    return __xml_parser__.from_string(xml, type)


def dataclass_to_xml(dc: dataclass) -> str:
    return serialize_dataclass(dc)



import json
import logging
import ssl
from http.server import BaseHTTPRequestHandler
from wsgiref.simple_server import WSGIRequestHandler

import OpenSSL
import werkzeug.exceptions
from flask import Flask, render_template, request, redirect, Response
from flask_sessions import Session
from werkzeug.serving import make_server, BaseWSGIServer

__all__ = ["build_server"]

# templates = Jinja2Templates(directory="templates")
# from IEEE2030_5.endpoints import dcap, IEEE2030_5Renderer
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.server.admin_endpoints import AdminEndpoints
from ieee_2030_5.server.server_endpoints import ServerEndpoints
from ieee_2030_5.server.server_constructs import get_groups, EndDevices

_log = logging.getLogger(__file__)


class PeerCertWSGIRequestHandler(werkzeug.serving.WSGIRequestHandler):
    """
    We subclass this class so that we can gain access to the connection
    property. self.connection is the underlying client socket. When a TLS
    connection is established, the underlying socket is an instance of
    SSLSocket, which in turn exposes the getpeercert() method.

    The output from that method is what we want to make available elsewhere
    in the application.
    """
    tlsrepo: TLSRepository

    def make_environ(self):
        """
        The superclass method develops the environ hash that eventually
        forms part of the Flask request object.

        We allow the superclass method to run first, then we insert the
        peer certificate into the hash. That exposes it to us later in
        the request variable that Flask provides
        """
        PeerCertWSGIRequestHandler.protocol_version = "HTTP/1.1"
        environ = super(PeerCertWSGIRequestHandler, self).make_environ()

        # Assume browser is being hit with things that start with /admin allow
        # a pass through from web (should be protected via auth but not right now)
        if environ['PATH_INFO'].startswith("/admin"):
            return environ

        try:
            x509_binary = self.connection.getpeercert(True)

            x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_ASN1, x509_binary)
            environ['ieee_2030_5_peercert'] = x509
            environ['ieee_2030_5_subject'] = x509.get_subject().CN
            environ['ieee_2030_5_serial_number'] = x509.get_serial_number()

        except OpenSSL.crypto.Error:
            environ['peercert'] = None

        # self.protocol_version = "HTTP/1.1"
        _log.debug("Making environ")
        return environ


# based on
# https://stackoverflow.com/questions/19459236/how-to-handle-413-request-entity-too-large-in-python-flask-server#:~:text=server%20MAY%20close%20the%20connection,client%20from%20continuing%20the%20request.&text=time%20the%20client%20MAY%20try,you%20the%20Broken%20pipe%20error.&text=Great%20than%20the%20application%20is%20acting%20correct.
def handle_chunking():
    """
    Sets the "wsgi.input_terminated" environment flag, thus enabling
    Werkzeug to pass chunked requests as streams.  The gunicorn server
    should set this, but it's not yet been implemented.
    """

    transfer_encoding = request.headers.get("Transfer-Encoding", None)
    if transfer_encoding == u"chunked":
        request.environ["wsgi.input_terminated"] = True


def before_request():
    _log.debug(f"HEADERS: {request.headers}")
    _log.debug(f"REQ_path: {request.path}")
    _log.debug(f"ARGS: {request.args}")
    _log.debug(f"DATA: {request.get_data()}")
    _log.debug(f"FORM: {request.form}")


def after_request(response: Response) -> Response:
    _log.debug(f"RESP:\n{response.get_data().decode('utf-8')}")
    return response


def build_server(config: ServerConfiguration,
                 tlsrepo: TLSRepository,
                 enddevices: EndDevices, **kwargs) -> BaseWSGIServer:
    app = Flask(__name__, template_folder="/repos/gridappsd-2030_5/templates")
    # Debug headers path and request arguments
    app.before_request(before_request)
    # Allows for larger data to be sent through because of chunking types.
    app.before_request(handle_chunking)
    app.after_request(after_request)

    # to establish an SSL socket we need the private key and certificate that
    # we want to serve to users.
    server_key_file = str(tlsrepo.server_key_file)
    server_cert_file = str(tlsrepo.server_cert_file)

    # in order to verify client certificates we need the certificate of the
    # CA that issued the client's certificate. In this example I have a
    # single certificate, but this could also be a bundle file.
    ca_cert = str(tlsrepo.ca_cert_file)

    # create_default_context establishes a new SSLContext object that
    # aligns with the purpose we provide as an argument. Here we provide
    # Purpose.CLIENT_AUTH, so the SSLContext is set up to handle validation
    # of client certificates.
    ssl_context = ssl.create_default_context(purpose=ssl.Purpose.CLIENT_AUTH, cafile=str(ca_cert))

    # load in the certificate and private key for our server to provide to clients.
    # force the client to provide a certificate.
    ssl_context.load_cert_chain(
        certfile=server_cert_file,
        keyfile=server_key_file,
        # password=app_key_password
    )
    # change this to ssl.CERT_REQUIRED during deployment.
    # TODO if required we have to have one all the time on the server.
    ssl_context.verify_mode = ssl.CERT_OPTIONAL  # ssl.CERT_REQUIRED

    ServerEndpoints(app, end_devices=enddevices, tls_repo=tlsrepo, config=config)
    AdminEndpoints(app, end_devices=enddevices, tls_repo=tlsrepo, config=config)

    # now we get into the regular Flask details, except we're passing in the peer certificate
    # as a variable to the template.
    @app.route('/')
    def root():
        return redirect("/admin/index.html")
        # cert = request.environ['peercert']
        # cert_data = f"{cert.get_subject()}"
        # return render_template("admin/index.html")
        # return render_template('helloworld.html', client_cert=request.environ['peercert'])

    @app.route("/admin/index.html")
    def admin_home():
        return render_template("admin/index.html")

    @app.route("/admin/clients")
    def admin_clients():
        clients = tlsrepo.client_list
        return json.dumps(clients)  # render_template("admin/clients.html", registered=clients, connected=[])

    @app.route("/admin/groups")
    def admin_groups():
        groups = get_groups()
        return render_template("admin/groups.html", groups=groups)

    @app.route("/admin/aggregators")
    def admin_aggregators():
        return Response("<h1>Aggregators</h1>")

    @app.route("/admin/routes")
    def admin_routes():
        routes = '<ul>'
        for p in app.url_map.iter_rules():
            routes += f"<li>{p.rule}</li>"
        routes += "</ul>"
        return Response(f"{routes}")

    try:
        host, port = config.server_hostname.split(":")
    except ValueError:
        # host and port not available
        host = config.server_hostname
        port = 8443
    # Add session support to the application.
    Session(app)

    return make_server(app=app,
                       host=host,
                       ssl_context=ssl_context,
                       request_handler=PeerCertWSGIRequestHandler,
                       port=port, **kwargs)



from __future__ import annotations

import json
from datetime import datetime, timedelta
import logging
from typing import Optional

import pytz
import tzlocal
from flask import Flask, Response, request
from werkzeug.exceptions import Forbidden

from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.data.indexer import get_href, get_all_filtered
from ieee_2030_5.models import Time, DERCurveList, DERProgramList
from ieee_2030_5.server.server_constructs import EndDevices
import ieee_2030_5.hrefs as hrefs
from ieee_2030_5.server.der_program import DERProgramRequests
from ieee_2030_5.server.edevrequests import EDevRequests, SDevRequests, DERRequests
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.server.uuid_handler import UUIDHandler

# module level instance of hrefs class.
from ieee_2030_5.server.usage_points import MUP, UTP
from ieee_2030_5.types_ import TimeOffsetType, format_time
from ieee_2030_5.utils import dataclass_to_xml

_log = logging.getLogger(__name__)


class Admin(RequestOp):
    def get(self):
        if not self.is_admin_client:
            raise Forbidden()
        return Response("We are able to do stuff here")

    def post(self):
        if not self.is_admin_client:
            raise Forbidden()
        return Response(json.dumps({'abc': 'def'}), headers={'Content-Type': 'application/json'})


class Dcap(RequestOp):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        return self.build_response_from_dataclass(self._end_devices.get_device_capability(self.lfid))


class TimeRequest(RequestOp):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        # TODO fix for new stuff.
        # local_tz = datetime.now().astimezone().tzinfo
        # now_local = datetime.now().replace(tzinfo=local_tz)

        now_utc = datetime.utcnow().replace(tzinfo=pytz.utc)
        # now_utc = pytz.utc.localize(datetime.utcnow())
        local_tz = pytz.timezone(tzlocal.get_localzone().zone)
        now_local = datetime.now().replace(tzinfo=local_tz)

        start_dst_utc, end_dst_utc = [
            dt for dt in local_tz._utc_transition_times if dt.year == now_local.year
        ]

        utc_offset = local_tz.utcoffset(start_dst_utc - timedelta(days=1))
        dst_offset = local_tz.utcoffset(start_dst_utc + timedelta(days=1)) - utc_offset
        local_but_utc = datetime.now().replace(tzinfo=pytz.utc)

        tm = Time(currentTime=format_time(now_utc),
                  dstEndTime=format_time(end_dst_utc.replace(tzinfo=pytz.utc)),
                  dstOffset=TimeOffsetType(int(dst_offset.total_seconds())),
                  localTime=format_time(local_but_utc),
                  quality=None,
                  tzOffset=TimeOffsetType(utc_offset.total_seconds()))

        return self.build_response_from_dataclass(tm)


class ServerList(RequestOp):
    def __init__(self, list_type: str, **kwargs):
        super().__init__(**kwargs)
        self._list_type = list_type

    def get(self) -> Response:
        response = None
        if self._list_type == 'EndDevice':
            response = self._end_devices.get_end_device_list(self.lfid)

        if response:
            response = dataclass_to_xml(response)

        return response


class ServerEndpoints:

    def __init__(self, app: Flask, end_devices: EndDevices, tls_repo: TLSRepository, config: ServerConfiguration):
        self.end_devices = end_devices
        self.config = config
        self.tls_repo = tls_repo
        self.mimetype = "text/xml"
        self.app: Flask = app

        _log.debug(f"Adding rule: {hrefs.uuid_gen} methods: {['GET']}")
        app.add_url_rule(hrefs.uuid_gen, view_func=self._generate_uuid)
        _log.debug(f"Adding rule: {hrefs.dcap} methods: {['GET']}")
        app.add_url_rule(hrefs.dcap, view_func=self._dcap)
        _log.debug(f"Adding rule: {hrefs.tm} methods: {['GET']}")
        app.add_url_rule(hrefs.tm, view_func=self._tm)
        _log.debug(f"Adding rule: {hrefs.sdev} methods: {['GET']}")
        app.add_url_rule(hrefs.sdev, view_func=self._sdev)
        _log.debug(f"Adding rule: {hrefs.derp} methods: {['GET']}")
        app.add_url_rule(hrefs.derp, view_func=self._derp)

        rulers = (
            (hrefs.der_urls, self._der),
            (hrefs.edev_urls, self._edev),
            (hrefs.mup_urls, self._mup),
            (hrefs.curve_urls, self._curves),
            (hrefs.program_urls, self._programs)
        )

        for endpoints, view_func in rulers:
            # Item should either be a single rule or a rule with a second element having the methods
            # in it.
            # edev = [
            #   /edev,
            #   (f"/edev/<int: index>", ["GET", "POST"])
            # ]
            for item in endpoints:
                try:
                    rule, methods = item
                except ValueError:
                    rule = item
                    methods = ["GET"]
                _log.debug(f"Adding rule: {rule} methods: {methods}")
                app.add_url_rule(rule, view_func=view_func, methods=methods)
        #
        # self.add_endpoint(hrefs.dcap, view_func=self._dcap)
        # self.add_endpoint(hrefs.edev, view_func=self._edev)
        # self.add_endpoint(hrefs.mup, view_func=self._mup, methods=['GET', 'POST'])
        # self.add_endpoint(hrefs.uuid_gen, view_func=self._generate_uuid)
        # app.add_url_rule(hrefs.rsps, view_func=None)
        # self.add_endpoint(hrefs.tm, view_func=self._tm)
        #
        # for index, ed in end_devices.all_end_devices.items():
        #     self.add_endpoint(hrefs.edev + f"/{index}", view_func=self._edev)
        #     self.add_endpoint(hrefs.mup + f"/{index}", view_func=self._mup)

    def _generate_uuid(self) -> Response:
        return Response(UUIDHandler().generate())
    #
    # def _admin(self) -> Response:
    #     return Admin(server_endpoints=self).execute()

    def _upt(self, index: Optional[int] = None) -> Response:
        return UTP(server_endpoints=self).execute(index=index)

    def _mup(self, index: Optional[int] = None) -> Response:
        return MUP(server_endpoints=self).execute(index=index)

    def _der(self, edev_id: int, id: Optional[int] = None) -> Response:
        return DERRequests(server_endpoints=self).execute(edev_id=edev_id, id=id)

    def _dcap(self) -> Response:
        return Dcap(server_endpoints=self).execute()

    def _edev(self, index: Optional[int] = None, category: Optional[str] = None) -> Response:
        return EDevRequests(server_endpoints=self).execute(index=index, category=category)

    def _sdev(self) -> Response:
        return SDevRequests(server_endpoints=self).execute()

    def _tm(self) -> Response:
        return TimeRequest(server_endpoints=self).execute()

    def _derp(self) -> Response:
        return DERProgramRequests(server_endpoints=self).execute()

    def _curves(self, index: Optional[int] = None) -> Response:
        if index is None:
            items = get_all_filtered(hrefs.curve)
            curve_list = DERCurveList(DERCurve=items, all=len(items), href=request.path, results=len(items))
            response = Response(dataclass_to_xml(curve_list))
        else:
            response = Response(dataclass_to_xml(get_href(request.path)))
        return response

    def _programs(self, index: Optional[int] = None) -> Response:
        if index is None:
            items = get_all_filtered(href_prefix=hrefs.program)
            program_list = DERProgramList(DERProgram=items, all=len(items), href=request.path, results=len(items))
            response = Response(dataclass_to_xml(program_list))
        else:
            response = Response(dataclass_to_xml(get_href(request.path)))
        return response



# -------------------------------------------------------------------------------
# Copyright (c) 2022, Battelle Memorial Institute All rights reserved.
# Battelle Memorial Institute (hereinafter Battelle) hereby grants permission to any person or entity
# lawfully obtaining a copy of this software and associated documentation files (hereinafter the
# Software) to redistribute and use the Software in source and binary forms, with or without modification.
# Such person or entity may use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and may permit others to do so, subject to the following conditions:
# Redistributions of source code must retain the above copyright notice, this list of conditions and the
# following disclaimers.
# Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
# the following disclaimer in the documentation and/or other materials provided with the distribution.
# Other than as used herein, neither the name Battelle Memorial Institute or Battelle may be used in any
# form whatsoever without the express written consent of Battelle.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
# BATTELLE OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
# OF THE POSSIBILITY OF SUCH DAMAGE.
# General disclaimer for use with OSS licenses
#
# This material was prepared as an account of work sponsored by an agency of the United States Government.
# Neither the United States Government nor the United States Department of Energy, nor Battelle, nor any
# of their employees, nor any jurisdiction or organization that has cooperated in the development of these
# materials, makes any warranty, express or implied, or assumes any legal liability or responsibility for
# the accuracy, completeness, or usefulness or any information, apparatus, product, software, or process
# disclosed, or represents that its use would not infringe privately owned rights.
#
# Reference herein to any specific commercial product, process, or service by trade name, trademark, manufacturer,
# or otherwise does not necessarily constitute or imply its endorsement, recommendation, or favoring by the United
# States Government or any agency thereof, or Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by BATTELLE for the
# UNITED STATES DEPARTMENT OF ENERGY under Contract DE-AC05-76RL01830
# -------------------------------------------------------------------------------

import os
import threading

from argparse import ArgumentParser
import logging
from pathlib import Path
import socket
import sys
from time import sleep

import yaml
from werkzeug.serving import BaseWSGIServer

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration
from ieee_2030_5.flask_server import build_server
from ieee_2030_5.server.server_constructs import initialize_2030_5


_log = logging.getLogger()


class ServerThread(threading.Thread):

    def __init__(self, server: BaseWSGIServer):
        threading.Thread.__init__(self, daemon=True)
        self.server = server

    def run(self):
        _log.info('starting server')
        self.server.serve_forever()

    def shutdown(self):
        _log.info("shutting down server")
        self.server.shutdown()


def get_tls_repository(cfg: ServerConfiguration, create_certificates_for_devices: bool = True) -> TLSRepository:
    tlsrepo = TLSRepository(cfg.tls_repository,
                            cfg.openssl_cnf,
                            cfg.server_hostname,
                            cfg.proxy_hostname,
                            clear=create_certificates_for_devices)
    if create_certificates_for_devices:
        already_represented = set()

        # registers the devices, but doesn't initialize_device the end devices here.
        for k in cfg.devices:
            if k in already_represented:
                _log.error(f"Already have {k.id} represented by {k.device_category_type}")
            else:
                already_represented.add(k)
                tlsrepo.create_cert(k.id)
    return tlsrepo


def _shutdown():
    make_stop_file()
    sleep(1)
    remove_stop_file()


def should_stop():
    return Path('server.stop').exists()


def make_stop_file():
    with open('server.stop', 'w') as w:
        pass


def remove_stop_file():
    pth = Path('server.stop')
    if pth.exists():
        os.remove(pth)


def _main():
    parser = ArgumentParser()

    parser.add_argument(dest="config", help="Configuration file for the server.")
    parser.add_argument("--no-validate",
                        action="store_true",
                        help="Allows faster startup since the resolving of addresses is not done!")
    parser.add_argument(
        "--no-create-certs",
        action="store_true",
        help="If specified certificates for for client and server will not be created.")
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Debug level of the server"
    )
    opts = parser.parse_args()

    logging_level = logging.DEBUG if opts.debug else logging.INFO
    logging.basicConfig(level=logging_level)

    _log = logging.getLogger(__name__)

    os.environ["IEEE_2030_5_CONFIG_FILE"] = str(
        Path(opts.config).expanduser().resolve(strict=True))

    cfg_dict = yaml.safe_load(Path(opts.config).expanduser().resolve(strict=True).read_text())

    config = ServerConfiguration(**cfg_dict)

    assert config.tls_repository
    assert len(config.devices) > 0
    assert config.server_hostname
    unknown = []
    # Only check for resolvability if not passed --no-validate
    if not opts.no_validate:
        _log.debug("Validating hostnames and/or ip of devices are resolvable.")
        for i in range(len(config.devices)):
            assert config.devices[i].hostname

            try:
                socket.gethostbyname(config.devices[i].hostname)
            except socket.gaierror:
                if hasattr(config.devices[i], "ip"):
                    try:
                        socket.gethostbyname(config.devices[i].ip)
                    except socket.gaierror:
                        unknown.append(config.devices[i].hostname)
                else:
                    unknown.append(config.devices[i].hostname)

    if unknown:
        _log.error("Couldn't resolve the following hostnames.")
        for host in unknown:
            _log.error(host)
        sys.exit(1)

    create_certs = not opts.no_create_certs
    tls_repo = get_tls_repository(config, create_certs)

    end_devices = initialize_2030_5(config, tls_repo)

    thread = None
    try:
        remove_stop_file()
        server = build_server(config, tls_repo, enddevices=end_devices)
        thread = ServerThread(server)
        thread.start()
        while not should_stop():
            sleep(0.5)
    except KeyboardInterrupt as ex:
        pass
    finally:
        if thread:
            thread.shutdown()
            thread.join()


if __name__ == '__main__':
    _main()



from __future__ import annotations

import inspect
import sys
from dataclasses import dataclass, field
import logging
from pathlib import Path
from typing import List, Literal, Union, Optional, Dict, Tuple

import yaml
from dataclasses_json import dataclass_json

__all__ = ["ServerConfiguration"]

try:
    from gridappsd.field_interface import MessageBusDefinition
except ImportError as ex:
    pass

from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.models import DeviceCategoryType, DERProgram, DERCurve, DefaultDERControl, DERControlBase, CurveData, \
    FunctionSetAssignments
from ieee_2030_5.types_ import Lfid

from ieee_2030_5.server.exceptions import NotFoundError


_log = logging.getLogger(__name__)


@dataclass
class DeviceConfiguration:
    id: str
    device_category_type: DeviceCategoryType
    pin: int
    hostname: str = None
    ip: str = None
    poll_rate: int = 900
    # TODO: Direct control means that only one FSA will be available to the client.
    direct_control: bool = True
    fsa_list: Optional[List[Dict]] = None
    DefaultDERControl: Optional[DefaultDERControl] = None

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __hash__(self):
        return self.id.__hash__()


@dataclass_json
@dataclass
class GridappsdConfiguration:
    field_bus_config: Optional[str] = None
    field_bus_def: Optional[MessageBusDefinition] = None
    feeder_id_file: Optional[str] = None
    feeder_id: Optional[str] = None
    simulation_id_file: Optional[str] = None
    simulation_id: Optional[str] = None


@dataclass
class ProgramList:
    name: str
    programs: List[DERProgram]


@dataclass
class ServerConfiguration:
    openssl_cnf: str
    # Can include ip address as well
    server_hostname: str
    server_mode: Union[Literal["enddevices_create_on_start"],
                       Literal["enddevices_register_access_only"]]
    devices: List[DeviceConfiguration]

    tls_repository: str
    openssl_cnf: str

    # map into program_lists array for programs for specific
    # named list.
    programs_map: Dict[str, int] = field(default_factory=dict)
    program_lists: List[ProgramList] = field(default_factory=list)
    fsa_list: List[FunctionSetAssignments] = field(default_factory=list)
    curve_list: List[DERCurve] = field(default_factory=list)

    proxy_hostname: Optional[str] = None
    gridappsd: Optional[GridappsdConfiguration] = None
    DefaultDERControl: Optional[DefaultDERControl] = None

    @classmethod
    def from_dict(cls, env):
        return cls(**{k: v for k, v in env.items() if k in inspect.signature(cls).parameters})

    def __post_init__(self):
        self.devices = [DeviceConfiguration.from_dict(x) for x in self.devices]
        for d in self.devices:
            d.device_category_type = eval(f"DeviceCategoryType.{d.device_category_type}")

        temp_program_list = self.program_lists
        if isinstance(self.program_lists, str):
            temp_program_list = yaml.safe_load(Path(self.program_lists).read_text())

        self.program_lists = []
        for program_list in temp_program_list['program_lists']:
            pl_obj = ProgramList(name=program_list["name"], programs=[])
            for inter_index, program_obj in enumerate(program_list['programs']):
                base = None
                if "DERControlBase" in program_obj:
                    base = DERControlBase()
                    for k in program_obj.get("DERControlBase"):
                        setattr(base, k, program_obj["DERControlBase"].get(k))
                    del program_obj["DERControlBase"]
                # TODO Do Something with base so we can retrieve it.
                # if base:
                #     self.DefaultDERControl = DefaultDERControl(DERControlBase=base)
                #     for k, v in program.items():
                #         setattr(self.DefaultDERControl, k, v)
                # else:
                program = DERProgram()

                for k, v in program_obj.items():
                    setattr(program, k, v)

                pl_obj.programs.append(program)
            self.program_lists.append(pl_obj)

        temp_curve_list = self.curve_list
        if isinstance(self.curve_list, str):
            temp_curve_list = yaml.safe_load(Path(self.curve_list).read_text())

        self.curve_list = []
        for item in temp_curve_list['curve_list']:
            curve_data: List[CurveData] = []
            for data in item.get('CurveData'):
                curve_data.append(CurveData(xvalue=data['xvalue'], yvalue=data['yvalue']))
            del item["CurveData"]

            curve = DERCurve(CurveData=curve_data)
            for k, v in item.items():
                setattr(curve, k, v)
            self.curve_list.append(curve)

        if self.gridappsd:
            self.gridappsd = GridappsdConfiguration.from_dict(self.gridappsd)
            if Path(self.gridappsd.feeder_id_file).exists():
                self.gridappsd.feeder_id = Path(self.gridappsd.feeder_id_file).read_text().strip()
            if Path(self.gridappsd.simulation_id_file).exists():
                self.gridappsd.simulation_id = Path(self.gridappsd.simulation_id_file).read_text().strip()

            if not self.gridappsd.feeder_id:
                raise ValueError("Feeder id from gridappsd not found in feeder_id_file nor was specified "
                                 "in gridappsd config section.")

            # TODO: This might not be the best place for this manipulation
            self.gridappsd.field_bus_def = MessageBusDefinition.load(self.gridappsd.field_bus_config)
            self.gridappsd.field_bus_def.id = self.gridappsd.feeder_id

            _log.info("Gridappsd Configuration For Simulation")
            _log.info(f"feeder id: {self.gridappsd.feeder_id}")
            if self.gridappsd.simulation_id:
                _log.info(f"simulation id: {self.gridappsd.simulation_id}")
            else:
                _log.info("no simulation id")
            _log.info("x" * 80)

        # if self.field_bus_config:
        #     self.field_bus_def = MessageBusDefinition.load(self.field_bus_config)

    def get_device_pin(self, lfid: Lfid, tls_repo: TLSRepository) -> int:
        for d in self.devices:
            test_lfid = tls_repo.lfdi(d.id)
            if test_lfid == int(lfid):
                return d.pin
        raise NotFoundError(f"The device_id: {lfid} was not found.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ieee_2030_5',
 'ieee_2030_5.client',
 'ieee_2030_5.data',
 'ieee_2030_5.models',
 'ieee_2030_5.persistance',
 'ieee_2030_5.server',
 'ieee_2030_5.simulation',
 'ieee_2030_5.types_',
 'ieee_2030_5.utils']

package_data = \
{'': ['*']}

install_requires = \
['Flask-Sessions>=0.1.5,<0.2.0',
 'Flask>=2.0.3,<3.0.0',
 'cryptography>=37.0.2,<38.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'gevent>=21.12.0,<22.0.0',
 'grequests>=0.6.0,<0.7.0',
 'gunicorn>=20.1.0,<21.0.0',
 'pickleDB>=0.9.2,<0.10.0',
 'pvlib>=0.9.0,<0.10.0',
 'pyOpenSSL>=22.0.0,<23.0.0',
 'simplekv>=0.14.1,<0.15.0',
 'trio>=0.21.0,<0.22.0',
 'tzlocal>=4.2,<5.0',
 'xsdata[cli]>=22.3,<23.0']

entry_points = \
{'console_scripts': ['2030_5_proxy = ieee_2030_5.basic_proxy:_main',
                     '2030_5_server = ieee_2030_5.__main__:_main',
                     '2030_5_shutdown = ieee_2030_5.__main__:_shutdown']}

setup_kwargs = {
    'name': 'gridappsd-2030-5',
    'version': '0.0.2a1',
    'description': '',
    'long_description': 'None',
    'author': 'C. Allwardt',
    'author_email': '3979063+craig8@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import logging
from pathlib import Path
from typing import List, Tuple

from cryptography import x509
from cryptography.hazmat.backends import default_backend

from ieee_2030_5.execute import execute_command

__all__ = ['TLSRepository']

from ieee_2030_5.types_ import PathStr, Lfid

_log = logging.getLogger(__name__)


class TLSRepository:

    def __init__(self, repo_dir: PathStr, openssl_cnffile_template: PathStr, serverhost: str, proxyhost: str = None,
                 clear=False):
        if isinstance(repo_dir, str):
            repo_dir = Path(repo_dir).expanduser().resolve()
        if isinstance(openssl_cnffile_template, str):
            openssl_cnffile_template = Path(openssl_cnffile_template).expanduser().resolve()
            if not openssl_cnffile_template.exists():
                raise ValueError(f"openssl_cnffile does not exist {openssl_cnffile_template}")
        self._repo_dir = repo_dir
        self._certs_dir = repo_dir.joinpath("certs")
        self._private_dir = repo_dir.joinpath("private")
        self._combined_dir = repo_dir.joinpath("combined")
        self._openssl_cnf_file = self._repo_dir.joinpath(openssl_cnffile_template.name)
        self._common_names = {serverhost: serverhost}
        if proxyhost:
            self._common_names[proxyhost] = proxyhost
        self._client_common_name_set = set()

        if not self._repo_dir.exists() or not self._certs_dir.exists() or \
                not self._private_dir.exists() or not self._combined_dir.exists():
            self._certs_dir.mkdir(parents=True)
            self._private_dir.mkdir(parents=True)
            self._combined_dir.mkdir(parents=True)

        index_txt = self._repo_dir.joinpath("index.txt")
        serial = self._repo_dir.joinpath("serial")
        if clear:
            for x in self._private_dir.iterdir():
                x.unlink()
            for x in self._certs_dir.iterdir():
                x.unlink()
            for x in self._combined_dir.iterdir():
                x.unlink()
            for x in self._repo_dir.iterdir():
                if x.is_file():
                    x.unlink()
            try:
                index_txt.unlink()
            except FileNotFoundError:
                pass
            try:
                serial.unlink()
            except FileNotFoundError:
                pass

        if not index_txt.exists():
            index_txt.write_text("")
        if not serial.exists():
            serial.write_text("01")

        new_contents = openssl_cnffile_template.read_text().replace("dir             = /home/gridappsd/tls",
                                                                    f"dir = {repo_dir}")
        self._openssl_cnf_file.write_text(new_contents)
        self._ca_key = self._private_dir.joinpath("ca.pem")
        self._ca_cert = self._certs_dir.joinpath("ca.crt")
        self._serverhost = serverhost
        self._proxyhost = proxyhost

        # Create a new ca key if not exists.
        if not Path(self._ca_key).exists():
            self.__create_ca__()
            __openssl_create_private_key__(self.__get_key_file__(self._serverhost))
            self.create_cert(self._serverhost, True)
            if proxyhost:
                self.create_cert(self._proxyhost, True)

        if not clear:
            for f in self._certs_dir.glob('*.crt'):
                f = Path(f)
                cn = self.get_common_name(f.stem)
                if cn not in ('ca', serverhost):
                    self._client_common_name_set.add(cn)

    def __create_ca__(self):
        __openssl_create_private_key__(self._ca_key)
        __openssl_create_ca_certificate__("ca", self._openssl_cnf_file, self._ca_key,
                                          self._ca_cert)
        __openssl_create_pkcs23_pem_and_cert__(self._ca_key, self._ca_cert,
                                               self.__get_combined_file__("ca"))

    def create_cert(self, common_name: str, as_server: bool = False):

        if not self.__get_key_file__(common_name).exists():
            __openssl_create_private_key__(self.__get_key_file__(common_name))
        __openssl_create_signed_certificate__(common_name, self._openssl_cnf_file, self._ca_key,
                                              self._ca_cert, self.__get_key_file__(common_name),
                                              self.__get_cert_file__(common_name), as_server)

        __openssl_create_pkcs23_pem_and_cert__(self.__get_key_file__(common_name), self.__get_cert_file__(common_name),
                                               self.__get_combined_file__(common_name))

        self._common_names[common_name] = common_name

    def lfdi(self, device_id: str) -> Lfid:
        fp = self.fingerprint(device_id, True)
        return Lfid(int(fp[:16], 16))

    def sfdi(self, device_id: str):
        lfdi_: int = self.lfdi(device_id)
        bit_left_truncation_len = 36
        # truncate the lFDI
        sfdi_no_sod_checksum = lfdi_ >> (lfdi_.bit_length() - bit_left_truncation_len)
        # calculate sum-of-digits checksum digit
        sod_checksum = 10 - sum([int(digit) for digit in str(sfdi_no_sod_checksum)]) % 10
        # right concatenate the checksum digit and return
        return str(sfdi_no_sod_checksum) + str(sod_checksum)

    def fingerprint(self, device_id: str, without_colan: bool = True) -> str:
        value = __openssl_fingerprint__(self.__get_cert_file__(device_id))
        if without_colan:
            value = value.replace(":", "")
        if "=" in value:
            value = value.split("=")[1]
        return value

    def get_common_name(self, device_id: str) -> x509:
        pem_data = Path(self.__get_cert_file__(device_id)).read_bytes()
        cert = x509.load_pem_x509_certificate(pem_data, default_backend())
        return cert.subject.get_attributes_for_oid(x509.oid.NameOID.COMMON_NAME)[0].value

    def get_file_pair(self, device_id: str) -> Tuple[str, str]:
        return (self.__get_cert_file__(device_id).as_posix(),
                self.__get_key_file__(device_id).as_posix())

    @property
    def client_list(self) -> List[str]:
        return list(self._client_common_name_set)

    @property
    def ca_key_file(self) -> Path:
        return self._ca_key

    @property
    def ca_cert_file(self) -> Path:
        return self._ca_cert

    @property
    def proxy_key_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_key_file__(self._proxyhost)

    @property
    def proxy_cert_file(self) -> Path:
        if not self._proxyhost:
            raise ValueError("No proxy host available")
        return self.__get_cert_file__(self._proxyhost)

    @property
    def server_key_file(self) -> Path:
        return self.__get_key_file__(self._serverhost)

    @property
    def server_cert_file(self) -> Path:
        return self.__get_cert_file__(self._serverhost)

    def __get_cert_file__(self, common_name: str) -> Path:
        return self._certs_dir.joinpath(f"{common_name}.crt")

    def __get_key_file__(self, common_name: str) -> Path:
        return self._private_dir.joinpath(f"{common_name}.pem")

    def __get_combined_file__(self, common_name: str) -> Path:
        return self._combined_dir.joinpath(f"{common_name}-combined.pem")


def __openssl_create_pkcs23_pem_and_cert__(private_key_file: Path, cert_file: Path, combined_file: Path):
    # openssl pkcs12 -export -in certificate.pem -inkey privatekey.pem -out cert-and-key.pfx
    tmpfile = Path("/tmp/tmp.p12")
    tmpfile2 = Path("/tmp/all.pem")
    tmpfile.unlink(missing_ok=True)
    cmd = ["openssl", "pkcs12", "-export", "-in", str(cert_file), "-inkey", str(private_key_file), "-out", str(tmpfile), "-passout", "pass:"]
    execute_command(cmd)

    # openssl pkcs12 -in path.p12 -out newfile.pem -nodes
    cmd = ["openssl", "pkcs12", "-in", str(tmpfile), "-out", str(tmpfile2), "-nodes", "-passin", "pass:"]
    #cmd = ["openssl", "pkcs12", "-in", str(tmpfile), "-out", str(combined_file), "-clcerts", "-nokeys", "-passin", "pass:"]
    # -clcerts -nokeys
    execute_command(cmd)

    with open(combined_file, "w") as fp:
        in_between = False
        for line in tmpfile2.read_text().split("\n"):
            if not in_between:
                if "BEGIN" in line:
                    fp.write(f"{line}\n")
                    in_between = True
            else:
                fp.write(f"{line}\n")
                if "END" in line:
                    in_between = False


def __openssl_create_private_key__(file_path: Path):
    # openssl ecparam -out private/ec-cakey.pem -name prime256v1 -genkey
    cmd = ["openssl", "ecparam", "-out", str(file_path), "-name", "prime256v1", "-genkey"]
    return execute_command(cmd)


def __openssl_create_ca_certificate__(common_name: str, opensslcnf: Path, private_key_file: Path,
                                      ca_cert_file: Path):
    # openssl req -new -x509 -days 3650 -config openssl.cnf \
    #   -extensions v3_ca -key private/ec-cakey.pem -out certs/ec-cacert.pem
    cmd = [
        "openssl", "req", "-new", "-x509", "-days", "3650", "-subj", f"/C=US/CN={common_name}",
        "-config",
        str(opensslcnf), "-extensions", "v3_ca", "-key",
        str(private_key_file), "-out",
        str(ca_cert_file)
    ]
    return execute_command(cmd)


def __openssl_create_csr__(common_name: str, opensslcnf: Path, private_key_file: Path,
                           server_csr_file: Path):
    subject_name = common_name.split(":")[0]
    # openssl req -new -key server.key -out server.csr -sha256
    cmd = [
        "openssl", "req", "-new", "-config",
        str(opensslcnf), "-subj", f"/C=US/CN={subject_name}", "-key",
        str(private_key_file), "-out",
        str(server_csr_file), "-sha256"
    ]
    return execute_command(cmd)


def __openssl_create_signed_certificate__(common_name: str,
                                          opensslcnf: Path,
                                          ca_key_file: Path,
                                          ca_cert_file: Path,
                                          private_key_file: Path,
                                          cert_file: Path,
                                          as_server: bool = False):
    subject_name = common_name.split(":")[0]
    csr_file = Path(f"/tmp/{common_name}")
    __openssl_create_csr__(common_name, opensslcnf, private_key_file, csr_file)
    # openssl ca -keyfile /root/tls/private/ec-cakey.pem -cert /root/tls/certs/ec-cacert.pem \
    #   -in server.csr -out server.crt -config /root/tls/openssl.cnf
    cmd = [
        "openssl",
        "ca",
        "-keyfile",
        str(ca_key_file),
        "-cert",
        str(ca_cert_file),
        "-subj",
        f"/C=US/CN={subject_name}",
        "-in",
        str(csr_file),
        "-out",
        str(cert_file),
        "-config",
        str(opensslcnf),
    # For no prompt use -batch
        "-batch"
    ]
    # if as_server:
    #     "-server"
    # print(" ".join(cmd))
    ret_value = execute_command(cmd)
    csr_file.unlink()
    return ret_value


def __openssl_fingerprint__(cert_file: Path, algorithm: str = "sha1"):

    if algorithm == "sha1":
        algorithm = "-sha1"
    else:
        raise NotImplementedError()

    cmd = ["openssl", "x509", "-in", str(cert_file), "-noout", "-fingerprint", algorithm]
    ret_value = execute_command(cmd)
    return ret_value



from typing import Optional

from flask import Response, request

from ieee_2030_5 import hrefs
from ieee_2030_5.models import Registration
from ieee_2030_5.server.base_request import RequestOp
from ieee_2030_5.utils import dataclass_to_xml


class DERRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self, edev_id: Optional[int] = None, id: Optional[int] = None) -> Response:
        return Response("Foo")


class EDevRequests(RequestOp):
    """
    Class supporting end devices and any of the subordinate calls to it.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self, index: Optional[int] = None, category: Optional[str] = None, id2: Optional[int] = None) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /edev
            /edev/0
            /edev/0/di
            /edev/0/reg

        """
        pth = request.environ['PATH_INFO']

        if not pth.startswith(hrefs.edev):
            raise ValueError(f"Invalid path for {self.__class__} {request.path}")

        # split returns a single value whether or not there was any characters found. if
        # this is the case then we want to return the list of the end devices.
        if index is None:
            retval = self._end_devices.get_end_device_list(self.lfid)
        else:
            # This should mean we have an index of an end device that we are going to return
            if category is None:
                retval = self._end_devices.get(index)
            else:

                if category == 'reg':
                    retval = self._end_devices.get_registration(index)

                elif category == 'fsa':
                    retval = self._end_devices.get_fsa(index)
                else:
                    raise NotImplementedError()

        return self.build_response_from_dataclass(retval)


class SDevRequests(RequestOp):
    """
    SelfDevice is an alias for the end device of a client.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get(self) -> Response:
        """
        Supports the get request for end_devices(EDev) and end_device_list_link.

        Paths:
            /sdev

        """
        end_device = self._end_devices.get_end_device_list(self.lfid).EndDevice[0]
        return self.build_response_from_dataclass(end_device)



from __future__ import annotations

import logging
from copy import copy, deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from enum import Flag, auto
from typing import Dict, Optional, List

import werkzeug.exceptions

from ieee_2030_5 import hrefs
from ieee_2030_5.certs import TLSRepository
from ieee_2030_5.config import ServerConfiguration, DeviceConfiguration, ProgramList
from ieee_2030_5.data.indexer import add_href, get_href
from ieee_2030_5.models import (
    EndDevice,
    DERProgram,
    DERProgramList,
    ActiveDERControlListLink,
    DefaultDERControlLink,
    DERControlListLink,
    DERCurveListLink, Registration, DERListLink, FunctionSetAssignmentsListLink, EndDeviceList, DeviceCategoryType,
    DefaultDERControl, RegistrationLink, ConfigurationLink, DeviceStatusLink, PowerStatusLink, DeviceInformationLink,
    LogEventListLink, SelfDeviceLink, EndDeviceListLink, DERProgramListLink, UsagePointListLink,
    MirrorUsagePointListLink, TimeLink, DeviceCapability, FunctionSetAssignments, DeviceInformation, DER,
    FunctionSetAssignmentsList)

from ieee_2030_5.server.uuid_handler import UUIDHandler
from ieee_2030_5.types_ import Lfid

_log = logging.getLogger(__name__)


class GroupLevel(Flag):
    """
    Each group is a construct of the layer the EndDevice is
    apart of.
    """
    System = auto()
    SubTransmission = auto()
    Substation = auto()
    Feeder = auto()
    Segment = auto()
    Transformer = auto()
    ServicePoint = auto()
    NonTopology = auto()


@dataclass
class Group:
    name: str
    description: str
    level: GroupLevel
    der_program: DERProgram
    _end_devices: Dict[bytes, EndDevice] = field(default_factory=dict)

    def add_end_device(self, end_device: EndDevice):
        self._end_devices[end_device.lFDI] = end_device

    def remove_end_device(self, end_device: EndDevice):
        self.remove_end_device_by_lfid(end_device.lFDI)

    def remove_end_device_by_lfid(self, lfid: bytes):
        del self._end_devices[lfid]

    def get_devices(self):
        return list(self._end_devices.values())


groups: Dict[GroupLevel, Group] = {}
der_programs: List[DERProgram] = []
uuid_handler: UUIDHandler = UUIDHandler()


def get_group(level: Optional[GroupLevel] = None, name: Optional[str] = None) -> Group:
    if not level and not name:
        raise ValueError("level or name must be specified to this function.")

    # if name exists then override the level with NonTopology
    if name:
        level = GroupLevel.NonTopology

    grp = groups.get(level)

    if not grp:
        raise ValueError(f"Invalid level specified {level}")

    if name is not None and level:
        for group in groups.values():
            if group.name == name:
                grp = group
                break

    return grp


def create_group(level: GroupLevel, name: Optional[str] = None) -> Group:
    if level is GroupLevel.NonTopology and not name:
        raise ValueError("NonTopology level must have a name associated with it")

    if level is not GroupLevel.NonTopology:
        mrid = "B" + str(level.name.__hash__())
        name = level.name
    else:
        mrid = "B" + str(name.__hash__())

    index = len(groups) + 1

    # TODO: Standardize urls so we can get them from a central spot.
    program_href = f"/sep2/A{index}/derp/1"
    program = DERProgram(mRID=mrid.encode('utf-8'),
                         description=name,
                         primacy=index * 10,
                         href=program_href)
    program.active_dercontrol_list_link = ActiveDERControlListLink(href=f"{program_href}/actderc")
    program.default_dercontrol_link = DefaultDERControlLink(href=f"{program_href}/dderc")
    program.dercontrol_list_link = DERControlListLink(href=f"{program_href}/derc")
    program.dercurve_list_link = DERCurveListLink(href=f"{program_href}/dc")

    if level not in groups:
        groups[level] = Group(level=level, name=name, description=name, der_program=program)

    der_programs.append(program)
    uuid_handler.add_known(mrid, program)


# Create all but the NonTopology group, which will get added
for _, lvl in enumerate(GroupLevel):
    create_group(lvl, name=lvl.name)

der_program_list = DERProgramList(DERProgram=der_programs)


def get_der_program_list():
    return der_program_list


def get_groups() -> Dict[GroupLevel, Group]:
    return groups


def initialize_2030_5(config: ServerConfiguration, tlsrepo: TLSRepository) -> EndDevices:
    """
    Initialize the 2030.5 server side.  After this function call the following items
    will be initialized.

    - Curve List
    - Program Lists
    If server_mode == "enddevices_create_on_start"
    - End Devices will be initialized and created
    """
    _log.debug("Initializing 2030.5")

    _log.debug("Update DERCurves' href property")
    # Create curves for der controls.
    for index, curve in enumerate(config.curve_list):
        curve.href = f"{hrefs.curve}/{index}"
        add_href(curve.href, curve)

    _log.debug("Update DERPrograms' adding links to the different program pieces.")
    # Initialize "global" DERPrograms href lists, including all of the different links to
    # locations for active, default, curve and control lists.
    for program_list in config.program_lists:
        for index, program in enumerate(program_list.programs):
            program.href = f"{hrefs.program}/{index}"
            program.ActiveDERControlListLink = ActiveDERControlListLink(href=f"{program.href}/actderc", all=0)
            program.DERCurveListLink = DERCurveListLink(href=f"{program.href}/dc", all=0)
            program.DefaultDERControlLink = DefaultDERControlLink(href=f"{program.href}/dderc")
            program.DERControlListLink = DERControlListLink(href=f"{program.href}/derc", all=0)

            add_href(program.href, program)

    _log.debug("Registering EndDevices")
    end_devices = EndDevices()
    if config.server_mode == "enddevices_create_on_start":
        program_list_names = [x.name for x in config.program_lists]
        for device_config in config.devices:
            end_device = end_devices.initialize_device(device_config=device_config,
                                                       lfid=tlsrepo.lfdi(device_config.id),
                                                       program_lists=config.program_lists)

            for fsa in device_config.fsa_list:
                print(fsa)
            print(end_devices.__all_end_devices__)

    return end_devices


@dataclass
class EndDeviceData:
    index: int
    mRID: str  # mrid for the device.
    end_device: EndDevice
    registration: Registration
    device_capability: DeviceCapability = None
    der_programs: Optional[List[DERProgram]] = field(default_factory=list)
    ders: Optional[List[DER]] = field(default_factory=list)
    function_set_assignments: Optional[List[FunctionSetAssignments]] = field(default_factory=list)
    device_information: Optional[DeviceInformation] = None


@dataclass
class EndDevices:
    """
    EndDevices contains the server side instances of an
    """
    __all_end_devices__: Dict[int, EndDeviceData] = field(default_factory=dict)
    _lfid_index_map: Dict[Lfid, int] = field(default_factory=dict)

    # only increasing device_numbers
    _last_device_number: int = field(default=-1)

    def initialize_groups(self):
        """
        Initialize groups so they are ready to go when registering devices for the
        different group levels of the system.
        """
        non_topo = get_group(level=GroupLevel.NonTopology)

        for index, indexer in self.__all_end_devices__.items():
            indexer.der_programs.append(non_topo.der_program)

    @property
    def num_devices(self) -> int:
        return len(self.__all_end_devices__)

    def get_end_devices(self) -> Dict[int, EndDevice]:
        devices: Dict[int, EndDevice] = {}
        for k, v in self.__all_end_devices__.items():
            devices[k] = copy(v.end_device)
        return devices

    def get_end_device_data(self, index: int) -> EndDeviceData:
        data = self.__all_end_devices__.get(index)
        if not data:
            raise werkzeug.exceptions.NotFound()

        return deepcopy(data)

    def get_fsa_list(self, lfid: Optional[Lfid] = None,
                     edevid: Optional[int] = None) -> List[FunctionSetAssignments] | []:
        if not ((lfid is not None) ^ edevid is not None):
            raise ValueError("Either lfid or edevid must be passed not both.")

        if lfid:
            indexer: EndDeviceData = self._lfid_index_map.get(lfid)
        else:
            indexer: EndDeviceData = self.__all_end_devices__.get(edevid)

        return indexer.function_set_assignments

    def get_device_capability(self, lfid: Lfid) -> DeviceCapability:
        if not isinstance(lfid, Lfid):
            lfid = Lfid(lfid)
        ed: EndDeviceData = self.__get_enddevicedata_by_lfid__(lfid)
        # if ed.device_capability is None:
        #     index = self._lfid_index_map[lfid].index
        #     sdev = SelfDeviceLink(href=hrefs.sdev)
        #     # TODO Add Aggregator for this
        #     edll = EndDeviceListLink(href=f"{hrefs.edev}", all=1)
        #     derp = DERProgramListLink(href=f"{hrefs.derp}", all=2)
        #     upt = UsagePointListLink(href=f"{hrefs.upt}", all=0)
        #     mup = MirrorUsagePointListLink(href=f"{hrefs.mup}", all=0)
        #     poll_rate = self.get_registration(index).pollRate
        #     timelink = TimeLink(href=f"{hrefs.tm}")
        #
        #     dc = DeviceCapability(
        #         href=hrefs.dcap,
        #         MirrorUsagePointListLink=mup,
        #         # SelfDeviceLink=sdev,
        #         EndDeviceListLink=edll,
        #         pollRate=poll_rate,
        #         TimeLink=timelink,
        #         UsagePointListLink=upt,
        #         DERProgramListLink=derp
        #     )
        #     self._lfid_index_map[lfid].device_capability = dc

        return ed.device_capability

    def get_device_by_index(self, index: int) -> EndDevice:
        return self.__get_enddevice_by_index__(index)

    def get_device_by_lfid(self, lfid: Lfid) -> EndDevice:
        index = self.__get_index_by_lfid__(lfid)
        return self.__get_enddevice_by_index__(index)

    def __get_enddevicedata_by_lfid__(self, lfid: Lfid):
        index = self.__get_index_by_lfid__(lfid)
        return self.__all_end_devices__[index]

    def __get_index_by_lfid__(self, lfid: Lfid):
        if not isinstance(lfid, Lfid):
            lfid = Lfid(lfid)

        index = self._lfid_index_map.get(lfid)
        if index is None:
            raise werkzeug.exceptions.NotFound()
        return index

    def __get_enddevice_by_index__(self, index: int) -> EndDevice:
        ed = self.__all_end_devices__.get(index)
        if ed is None:
            raise werkzeug.exceptions.NotFound()
        return ed.end_device

    def initialize_device(self, device_config: DeviceConfiguration, lfid: Lfid,
                          program_lists: List[ProgramList]) -> EndDevice:
        ts = int(round(datetime.utcnow().timestamp()))
        self._last_device_number += 1
        new_dev_number = self._last_device_number

        # Manage links to different resources for the device.
        reg_link_href = hrefs.build_edev_registration_link(new_dev_number)
        reg_link = RegistrationLink(href=reg_link_href)

        cfg_link_href = hrefs.build_edev_config_link(new_dev_number)
        cfg_link = ConfigurationLink(cfg_link_href)

        dev_status_link_href = hrefs.build_edev_status_link(new_dev_number)
        dev_status_link = DeviceStatusLink(href=dev_status_link_href)

        power_status_link_href = hrefs.build_edev_power_status_link(new_dev_number)
        power_status_link = PowerStatusLink(href=power_status_link_href)

        # file_status_link = FileStatusLink(href=hrefs.edev_file_status_fmt.format(
        #     index=new_dev_number))
        dev_info_link_href = hrefs.build_edev_info_link(new_dev_number)
        dev_info_link = DeviceInformationLink(href=dev_info_link_href)

        # sub_list_link = SubscriptionListLink(href=hrefs.edev_sub_list_fmt.format(
        #     index=new_dev_number))
        l_fid_bytes = str(lfid).encode('utf-8')

        base_edev_single = hrefs.extend_url(hrefs.edev, new_dev_number)
        der_list_link_href = hrefs.build_der_link(new_dev_number)
        der_list_link = DERListLink(href=der_list_link_href)

        fsa_list_link_href = hrefs.extend_url(base_edev_single, suffix="fsa")
        fsa_list_link = FunctionSetAssignmentsListLink(href=fsa_list_link_href)

        log_event_list_link_href = hrefs.extend_url(base_edev_single, suffix="log")
        log_event_list_link = LogEventListLink(href=log_event_list_link_href)

        time_link_href = hrefs.tm
        time_link = TimeLink(href=time_link_href)

        end_device_href = f"{hrefs.edev}/{new_dev_number}"
        end_device_list_link = EndDeviceListLink(href=hrefs.edev)

        changed_time = datetime.now()
        changed_time.replace(microsecond=0)

        program_list_names = [x.name for x in program_lists]
        found_fsa_item = set()
        fsa_items: List[FunctionSetAssignments] = []
        for fsa in device_config.fsa_list:
            found_program = None
            for fsa_name in fsa['program_lists']:
                for program_list in program_lists:
                    if program_list.name == fsa_name:
                        found_program = program_list
            if found_program is None:
                raise ValueError(f"Invalid fsa: {fsa} not found in program_lists.  Check configuration.")

            fsa_items.append(FunctionSetAssignments(mRID=fsa['mRID'], description=fsa['description'],
                                                    href=hrefs.extend_url(fsa_list_link_href, len(fsa_items))))

        device_capability_link = hrefs.dcap
        device_capability = DeviceCapability(href=device_capability_link, TimeLink=time_link,
                                             EndDeviceListLink=end_device_list_link)
        end_device = EndDevice(deviceCategory=device_config.device_category_type.value,
                               lFDI=l_fid_bytes,
                               RegistrationLink=reg_link,
                               DeviceStatusLink=dev_status_link,
                               ConfigurationLink=cfg_link,
                               PowerStatusLink=power_status_link,
                               DeviceInformationLink=dev_info_link,
                               # TODO: Do actual sfid rather than lfid.
                               sFDI=lfid,
                               # file_status_link=file_status_link,
                               # subscription_list_link=sub_list_link,
                               href=end_device_href,
                               # DERListLink=der_list_link,
                               FunctionSetAssignmentsListLink=fsa_list_link,
                               LogEventListLink=log_event_list_link,
                               enabled=True,
                               changedTime=int(changed_time.timestamp()))

        add_href(end_device_href, end_device)

        registration = Registration(dateTimeRegistered=ts, pollRate=device_config.poll_rate, pIN=device_config.pin)
        add_href(reg_link_href, registration)
        edd = EndDeviceData(index=new_dev_number, mRID=device_config.id,
                            end_device=end_device, registration=registration,
                            function_set_assignments=fsa_items,
                            device_capability=device_capability)
        self.__all_end_devices__[new_dev_number] = edd
        self._lfid_index_map[lfid] = new_dev_number
        return get_href(end_device_href)

    def get(self, index: int) -> EndDevice:
        return self.__all_end_devices__[index].end_device

    def get_registration(self, index: int) -> Registration:
        return self.__all_end_devices__[index].registration

    def get_der_list(self, index: int) -> DERListLink:
        return self.__all_end_devices__[index].end_device.DERListLink

    def get_fsa(self, index: int) -> FunctionSetAssignmentsListLink:
        return self.__all_end_devices__[index].end_device.FunctionSetAssignmentsListLink

    def get_end_device_list(self, lfid: Lfid, start: int = 0, length: int = 1) -> EndDeviceList:
        ed = self.get_device_by_lfid(lfid)
        if DeviceCategoryType(ed.deviceCategory) == DeviceCategoryType.AGGREGATOR:
            devices = [x.end_device for x in self.__all_end_devices__.values()]
        else:
            devices = [ed]

        # TODO Handle start, length list things.
        dl = EndDeviceList(EndDevice=devices, all=len(devices), results=len(devices), href=hrefs.edev, pollRate=900)
        return dl


if __name__ == '__main__':
    print(get_groups())
    for x in der_program_list.DERProgram:
        print(x.href)

    for x, v in get_groups().items():
        print(x)
        print(v)

    create_group(name="foo")
    g = get_group(name="foo")
    assert GroupLevel.NonTopology == g.level
    assert "foo" == g.name
